package ifresco.fcpro.client;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.HttpEntity;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.Credentials;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.conn.ssl.X509HostnameVerifier;
import org.apache.http.entity.ContentProducer;
import org.apache.http.entity.EntityTemplate;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicHeader;
import org.apache.http.params.CoreConnectionPNames;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

/**
 * Client to access remote AutoOCR services.
 *
 * <pre>
 * {@code
 * // Initialize client.
 * FCproClient client = new FCproClient();
 * client.setEndpoint("https://example.com/FileConverterProREST");
 * client.setUser("admin");
 * client.setPassword("secret");
 * client.setApiHash(123456);
 *
 * // Get server state.
 * logger.debug("==== Server State ====");
 * logger.debug("getServerHash():\t" + client.getServerHash());
 * logger.debug("getServerVersion():\t" + client.getServerVersion());
 * logger.debug("getAvgSecPerPage():\t" + client.getAvgSecPerPage());
 * logger.debug("getNrOfDocumentsInQueue():\t" + client.getNrOfDocumentsInQueue());
 * logger.debug("getNrOfPagesInQueue():\t" + client.getNrOfPagesInQueue());
 * logger.debug("getAvailablePages():\t" + client.getAvailablePages());
 *
 * // Get available settings/profiles.
 * logger.debug("==== getSettings ====");
 * FCproSetting[] settings = client.getSettings();
 * for (FCproSetting setting : settings) {
 *     logger.debug(setting.getSettingsName());
 * }
 *
 * // Upload file for conversion.
 * FCproJob upload = client.upload(new File("/tmp/example.pdf"), "application/pdf", "DE-DE", "example-label");
 * logger.debug(upload.getGuid());
 * // Wait up to 60 seconds for completion, check every two seconds.
 * FCproStatus completionStatus = client.waitForCompletion(upload.getGuid(), 60_000, 2_000);
 * logger.debug(completionStatus.toString());
 * if (completionStatus == FCproStatus.CONVERTED) {
 *     client.download(upload.getGuid(), new File("/tmp/example-converted.pdf");
 * }
 * }
 * </pre>
 *
 * @since 1.0
 */
public class FCproClient {

    private final Log logger = LogFactory.getLog(FCproClient.class);
    /**
     * REST endpoint, in the form of http://hostname.com/servicename.
     */
    private String endpoint;
    /**
     * Username used to authenticate against the service.
     */
    private String user;
    /**
     * The password used to authenticate against the service.
     */
    private String password;
    /**
     * Encoding used for Basic Authentication credentials.
     */
    private String credentialsCharset = "UTF-8";

    /**
     * Encoding used for url parameters.
     */
    private String urlEncodingCharset = "UTF-8";

    /**
     * HTTP connection timeout. The value 0 (default) means infinite. This is
     * also used as a transfer timeout.
     */
    private int connectionTimeout = 0;

    /**
     * Hash that allows you to work with the REST-API from AutoOCR
     */
    private int apiHash;

    /**
     * State variable, to record a successful validation of the api hash key.
     */
    private boolean apiKeyValidatedSuccessfully = false;

    /**
     * Used to parse ISO dates.
     */
    private final SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssz");

    /**
     * Quering for all users.
     *
     * @return List of all Users
     */
    public List<FCproUser> getUsers() {
        List<FCproUser> users = new ArrayList<>();

        try {
            String methodUri = "/Admin/GetUsers";
            JSONObject json = doSimpleJsonGet(methodUri);
            JSONArray arr = json.getJSONArray("GetUsersResult");

            for (int i = 0; i < arr.length(); i++) {

                JSONObject j = arr.getJSONObject(i);
                String userName = j.getString("UserName");
                int permission = j.optInt("Rights");
                FCproUser localUser = new FCproUser();
                localUser.setUserName(userName);
                localUser.setPermissions(permission);
                users.add(localUser);

            }
            return users;

        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }

    }

    /**
     * Asking for AutoOCR-Server status.
     *
     * @return returns true if server is running
     */
    public boolean isStarted() {
        try {
            String methodUri = "/Admin/IsStarted";
            JSONObject json = doSimpleJsonGet(methodUri);
            boolean isStarted = json.getBoolean("IsStartedResult");

            return isStarted;

        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }

    }

    /**
     * Starts the AutoOCR-Server.
     */
    public void startServer() {

        String methodUri = "/Admin/Start";
        doSimpleJsonGet(methodUri);
    }

    /**
     * Stops the AutoOCR-Server.
     */
    public void stopServer() {
        String methodUri = "/Admin/Stop";
        doSimpleJsonGet(methodUri);
    }

    /**
     * Transformation speed information.
     *
     * @return the average number of secods needed for a page.
     */
    public int getAvgSecPerPage() {
        try {
            String methodUri = "/GetAvgSecPerPage";
            JSONObject json = doSimpleJsonGet(methodUri);
            return json.getInt("GetAvgSecPerPageResult");
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Returns the total number of documents currently stored in the queue.
     *
     * @return the total number of documents in the queue.
     */
    public int getNrOfDocumentsInQueue() {
        try {
            String methodUri = "/GetNrOfDocumentsInQueue";
            JSONObject json = doSimpleJsonGet(methodUri);
            return json.getInt("GetNrOfDocumentsInQueueResult");
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Returns the number of total pages currently stored in the queue.
     *
     * @return the total number of pages in the queue.
     */
    public int getNrOfPagesInQueue() {
        try {
            String methodUri = "/GetNrOfPagesInQueue";
            JSONObject json = doSimpleJsonGet(methodUri);
            return json.getInt("GetNrOfPagesInQueueResult");
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Some OCR engines maintain a page limit, which can be queried through this
     * method.
     *
     * @return the number of available pages.
     */
    public int getAvailablePages() {
        try {
            String methodUri = "/GetAvailablePages";
            JSONObject json = doSimpleJsonGet(methodUri);
            return json.getInt("GetAvailablePagesResult");
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Removes a job from the queue.
     *
     * @param guid the job GUID.
     */
    public void removeJob(String guid) {
        try {
            String methodUri = "/RemoveJob?jobID=" + URLEncoder.encode(guid, urlEncodingCharset);
            doSimpleJsonGet(methodUri);
        } catch (UnsupportedEncodingException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Cancels a job.
     *
     * @param guid the job GUID.
     */
    public void cancelJob(String guid) {
        try {
            String methodUri = "/CancelJob?jobID=" + URLEncoder.encode(guid, urlEncodingCharset);
            doSimpleJsonGet(methodUri);
        } catch (UnsupportedEncodingException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Delets all jobs from the queue.
     */
    public void deleteAllJobs() {
        String methodUri = "/DeleteAllJobs";
        if (logger.isDebugEnabled()) {
            logger.debug(methodUri);
        }
        doSimpleJsonGet(methodUri);
    }

    /**
     * Deletes all jobs older than a given number of days.
     *
     * @param days the number of days
     */
    public void deleteJobsOlderThanDays(int days) {
        String methodUri = "/DeleteJobs?days=" + days;
        if (logger.isDebugEnabled()) {
            logger.debug(methodUri);
        }
        doSimpleJsonGet(methodUri);
    }

    /**
     * Returns the file extension of the transformed file at the given index.
     *
     * @param guid the job GUID
     * @param index the index
     * @return the file extension of the job
     */
    public String getResultExt(String guid, int index) {
        try {
            String methodUri = "/GetResultExt?jobID=" + URLEncoder.encode(guid, urlEncodingCharset) + "&index=" + index;
            JSONObject json = doSimpleJsonGet(methodUri);
            return json.getString("GetResultExtResult");
        } catch (UnsupportedEncodingException | JSONException e) {
            throw new GenericFCproClientException(e);
        }

    }

    /**
     * Returns the file mimetype of the transformed file at the given index.
     *
     * @param guid the job GUID
     * @param index the index
     * @return the file mimetype
     */
    public String getResultMimetype(String guid, int index) {
        try {
            String methodUri = "/GetResultMimeType?jobID=" + URLEncoder.encode(guid, urlEncodingCharset) + "&index=" + index;
            JSONObject json = doSimpleJsonGet(methodUri);
            return json.getString("GetResultMimeTypeResult");
        } catch (UnsupportedEncodingException | JSONException e) {
            throw new GenericFCproClientException(e);
        }

    }

    /**
     * Queries the number of transformation results. Only valid if the job is in
     * state CONVERTED.
     *
     * @param guid the job GUID
     * @return the number of transformation results
     */
    public int getResultCount(String guid) {
        try {
            String methodUri = "/GetResultCount?jobID=" + URLEncoder.encode(guid, urlEncodingCharset);
            JSONObject json = doSimpleJsonGet(methodUri);
            return json.getInt("GetResultCountResult");
        } catch (JSONException | UnsupportedEncodingException e) {
            throw new GenericFCproClientException(e);
        }

    }

    /**
     * Getting the remote API key from AutoOCR-Server.
     *
     * @return the API key hash
     */
    public int getServerHash() {
        try {
            String methodUri = "/GetServerHash";
            JSONObject json = doSimpleJsonGet(methodUri);
            if (logger.isDebugEnabled()) {
                logger.debug("getServerHash, JSON=" + json.toString(2));
            }
            return json.getInt("GetServerHashResult");
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Getting version of AutoOCR-Server.
     *
     * @return the server version
     */
    public String getServerVersion() {
        try {
            String methodUri = "/GetServerVersion";
            JSONObject json = doSimpleJsonGet(methodUri);
            if (logger.isDebugEnabled()) {
                logger.debug("getServerHash, JSON=" + json.toString(2));
            }
            return json.getString("GetServerVersionResult");
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Waits for the completion of a transformation. The backend ocr service is
     * polled to monitor the completion of the job. If the job did not complete
     * in time, a {@link TransformationTimedOutException} is thrown. If it
     * completes in time, the resulting job status is returned and should be
     * checked for errors.
     *
     * @param guid the job GUID
     * @param timeout milliseconds to wait for completion before aborting with
     * {@link TransformationTimedOutException}
     * @param sleeptime time in milliseconds to wait between polls
     * @throws TransformationTimedOutException if the transformation did not
     * complete in timeout time.
     * @return the job status
     */
    public FCproStatus waitForCompletion(String guid, long timeout, long sleeptime) {
        long start = System.currentTimeMillis();
        long diff = 0;
        do {
            if (diff > 0) {
                try {
                    Thread.sleep(sleeptime);
                } catch (InterruptedException e) {
                    throw new GenericFCproClientException(e);
                }
            }

            FCproStatus status = getStatus(guid);
            switch (status) {
                case UPLOADED:
                case CONVERTING:
                    break;
                default:
                    if (logger.isDebugEnabled()) {
                        logger.debug("transformation job finished: time=" + (System.currentTimeMillis() - start) / 1000 + "sec, guid=" + guid + ", result="
                                + status);
                    }
                    return status;
            }
            long now = System.currentTimeMillis();
            diff = now - start;
        } while (diff < timeout);

        throw new TransformationTimedOutException("Transformation did not complete: guid=" + guid + ", timeout=" + timeout + "msecs");
    }

    /**
     * Directly query for a job status.
     *
     * @param guid the job GUID
     * @return the job status
     */
    public FCproStatus getStatus(String guid) {
        try {
            String methodUri = "/GetStatus?jobID=" + URLEncoder.encode(guid, urlEncodingCharset);
            JSONObject json = doSimpleJsonGet(methodUri);
            return FCproStatus.getStatusFromValue(json.getInt("GetStatusResult"));
        } catch (JSONException | UnsupportedEncodingException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Query for a job. If there is no job for the guid an empty
     * {@link FCproJob} object is returned with
     * {@code status == FCproStatus.INVALID}.
     *
     * @param guid the job GUID
     * @return the job
     */
    public FCproJob getJob(String guid) {

        try {
            String methodUri = "/GetJob?jobID=" + URLEncoder.encode(guid, urlEncodingCharset);
            JSONObject json = doSimpleJsonGet(methodUri);
            FCproJob job = jsonToJob(json.getJSONObject("GetJobResult"));
            return job;
        } catch (UnsupportedEncodingException | JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Query for all jobs from a user, having the specified state.
     *
     * @param state the state
     * @return a list of all jobs having the state
     */
    public List<FCproJob> getOwnJobs(FCproStatus state) {
        try {
            List<FCproJob> list = new ArrayList<>();
            int int_state = FCproStatus.toIntValue(state);
            String str_state = Integer.toString(int_state);
            String methodUri = "/GetOwnJobs?status=" + URLEncoder.encode(str_state, urlEncodingCharset);
            JSONObject json = doSimpleJsonGet(methodUri);
            JSONArray arr = json.getJSONArray("GetOwnJobsResult");
            for (int i = 0; i < arr.length(); i++) {
                FCproJob job = jsonToJob(arr.getJSONObject(i));
                list.add(job);

            }
            return list;
        } catch (UnsupportedEncodingException | JSONException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Query for all jobs filtered by state state.
     *
     * @param state the stats
     * @return a list of all jobs having the state
     */
    public List<FCproJob> getAllJobs(FCproStatus state) {
        if (FCproStatus.ALL.equals(state)) {
            return getAllJobs();
        }
        try {
            int int_state = FCproStatus.toIntValue(state);
            String str_state = Integer.toString(int_state);
            List<FCproJob> list = new ArrayList<>();
            String methodUri = "/GetAllJobs?status=" + URLEncoder.encode(str_state, urlEncodingCharset);
            JSONObject json = doSimpleJsonGet(methodUri);
            JSONArray arr = json.getJSONArray("GetAllJobsResult");
            for (int i = 0; i < arr.length(); i++) {
                FCproJob job = jsonToJob(arr.getJSONObject(i));
                list.add(job);

            }
            return list;
        } catch (UnsupportedEncodingException | JSONException e) {
            throw new GenericFCproClientException(e);
        }

    }

    /**
     * Query for all jobs but without filtering the results.
     *
     * @return a list of all jobs
     */
    public List<FCproJob> getAllJobs() {
        List<FCproJob> list = new ArrayList<>();

        list.addAll(getAllJobs(FCproStatus.CONVERSION_ERROR));
        list.addAll(getAllJobs(FCproStatus.CONVERTED));
        list.addAll(getAllJobs(FCproStatus.CONVERTING));
        list.addAll(getAllJobs(FCproStatus.CREATED));
        list.addAll(getAllJobs(FCproStatus.DOWNLOADED));
        list.addAll(getAllJobs(FCproStatus.EXPIRED));
        list.addAll(getAllJobs(FCproStatus.INVALID));
        list.addAll(getAllJobs(FCproStatus.UPLOADED));
        list.addAll(getAllJobs(FCproStatus.CANCELED));
        return list;

    }

    /**
     * Get a filtered job enumeration. Used for paging requests.
     *
     * @param pageSize the page size
     * @param label the label to filter for
     * @param user the user to filter for
     * @param status the status to filter for
     * @param createdFromTo the created date range to filter for
     * @param convertedFromTo the converted date range to filter for
     * @param sortColumns the columns to sort the result
     * @return the job enumeration
     */
    public FCproEnumJobsResult enumJobs(int pageSize, String label, FCproUser user, FCproStatus status, String createdFromTo, String convertedFromTo,
            String sortColumns) {
        FCproEnumJobsResult result = new FCproEnumJobsResult();

        try {
            StringBuilder sb = new StringBuilder("/EnumJobs?");
            sb.append("pageSize=").append(pageSize);
            sb.append("&label=");
            if (label != null) {
                sb.append(URLEncoder.encode(label, urlEncodingCharset));
            } else {
                sb.append(URLEncoder.encode("*", urlEncodingCharset));
            }
            sb.append("&userName=");
            if (user != null && user.getUserName() != null) {
                sb.append(URLEncoder.encode(user.getUserName(), urlEncodingCharset));
            } else {
                sb.append(URLEncoder.encode("*", urlEncodingCharset));
            }
            sb.append("&status=").append(FCproStatus.toIntValue(status));
            sb.append("&createdFromTo=");
            if (createdFromTo != null) {
                sb.append(URLEncoder.encode(createdFromTo, urlEncodingCharset));
            }
            sb.append("&convertedFromTo=");
            if (convertedFromTo != null) {
                sb.append(URLEncoder.encode(convertedFromTo, urlEncodingCharset));
            }
            if (null != sortColumns) // Map YUI columns to AutoOCR columns.
            {
                switch (sortColumns) {
                    case "label ASC":
                        sortColumns = "JobLabel ASC";
                        break;
                    case "label DESC":
                        sortColumns = "JobLabel DESC";
                        break;
                    case "user ASC":
                        sortColumns = "UserName ASC";
                        break;
                    case "user DESC":
                        sortColumns = "UserName DESC";
                        break;
                    case "status ASC":
                        sortColumns = "StatusCode ASC";
                        break;
                    case "status DESC":
                        sortColumns = "StatusCode DESC";
                        break;
                    case "created ASC":
                        sortColumns = "DateCreated ASC";
                        break;
                    case "created DESC":
                        sortColumns = "DateCreated DESC";
                        break;
                    case "converted ASC":
                        sortColumns = "DateConverted ASC";
                        break;
                    case "converted DESC":
                        sortColumns = "DateConverted DESC";
                        break;
                    default:
                        sortColumns = "DateCreated ASC";
                        break;
                }
            }
            sb.append("&sortColumns=").append(URLEncoder.encode(sortColumns, urlEncodingCharset));

            String methodUri = sb.toString();
            if (logger.isDebugEnabled()) {
                logger.debug("methodUri: " + methodUri);
            }
            JSONObject json = doSimpleJsonGet(methodUri);
            result = jsonToEnumJobResult(json.getJSONObject("EnumJobsResult"));
        } catch (JSONException | UnsupportedEncodingException e) {
            throw new GenericFCproClientException(e);
        }

        return result;
    }

    /**
     * Get a page of jobs from a job enumeration.
     *
     * @param enumID the enumeration ID
     * @param pageIndex the requested page of the enumeration
     * @return the page of jobs
     */
    public List<FCproJob> enumJobsGetPage(int enumID, int pageIndex) {
        List<FCproJob> list = new ArrayList<>();

        try {
            StringBuilder sb = new StringBuilder("/EnumJobsGetPage?");
            sb.append("enumID=").append(enumID);
            sb.append("&pageIndex=").append(pageIndex);
            String methodUri = sb.toString();
            JSONObject json = doSimpleJsonGet(methodUri);

            JSONArray arr = json.getJSONArray("EnumJobsGetPageResult");
            for (int i = 0; i < arr.length(); i++) {
                FCproJob job = jsonToJob(arr.getJSONObject(i));
                list.add(job);
            }
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }

        return list;
    }

    /**
     * Query AutoOCR server settings.
     *
     * @return an array of settings
     */
    public FCproSetting[] getSettings() {
        Set<FCproSetting> settings = new HashSet<>();
        try {
            String methodUri = "/GetSettingsCollection";
            JSONObject json = doSimpleJsonGet(methodUri);
            JSONArray list = json.getJSONArray("GetSettingsCollectionResult");
            for (int i = 0; i < list.length(); i++) {
                String settingsName = list.getJSONObject(i).getString("SettingsName");
                FCproSetting setting = new FCproSetting(settingsName);
                setting.setDescription(list.getJSONObject(i).getString("Description"));
                setting.setEngineName(list.getJSONObject(i).getString("EngineName"));
                JSONArray addOns = list.getJSONObject(i).getJSONArray("AdditionalOutput");

                for (int k = 0; k < addOns.length(); k++) {
                    String format = addOns.getString(k);
                    if (StringUtils.isNotBlank(format)) {
                        setting.getAdditonalOutput().add(format);
                    }
                }

                JSONArray inputFormats = list.getJSONObject(i).getJSONArray("InputFormats");
                List<String> iF = jsonArr2List(inputFormats);
                if (!iF.contains("pdf")) {
                    iF.add(0, "pdf");
                }
                setting.setInputFormats(iF);

                JSONArray inputFormatsMT = list.getJSONObject(i).getJSONArray("InputFormatsMT");
                List<String> iFMT = jsonArr2List(inputFormatsMT);
                if (!iFMT.contains("application/pdf")) {
                    iFMT.add(0, "application/pdf");
                }
                setting.setInputFormatsMT(iFMT);

                JSONArray outputFormats = list.getJSONObject(i).getJSONArray("OutputFormats");
                List<String> oF = jsonArr2List(outputFormats);
                setting.setOutputFormats(oF);
                if (!oF.contains("pdf")) {
                    oF.add(0, "pdf");
                }

                JSONArray outputFormatsMT = list.getJSONObject(i).getJSONArray("OutputFormatsMT");
                List<String> oFMT = jsonArr2List(outputFormatsMT);
                setting.setOutputFormatsMT(oFMT);
                if (!oFMT.contains("application/pdf")) {
                    oFMT.add(0, "application/pdf");
                }

                settings.add(setting);

            }

        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }

        return settings.toArray(new FCproSetting[]{});
    }

    private List<String> jsonArr2List(JSONArray arr) throws JSONException {

        List<String> lst = new ArrayList<>();
        for (int k = 0; k < arr.length(); k++) {
            String format = arr.getString(k);
            if (StringUtils.isNotBlank(format)) {
                lst.add(format);
            }
        }

        return lst;

    }

    /**
     * Upload and create a job from a local file. Single output format version.
     *
     * @param file the file to upload
     * @param ext the target transformation format as autoocr extension or
     * mimetype
     * @param profile the server profile to use
     * @param label description field for document name or something like that
     * @return the job created by the upload
     */
    public FCproJob upload(File file, String ext, String profile, String label) {
        return upload(file, ext, profile, null, label);
    }

    /**
     * Upload and create a job from a local file.
     *
     * @param file the file to upload
     * @param ext the target transformation format as autoocr extension or
     * mimetype
     * @param profile the server profile to use
     * @param additionals a list of file formats to be created as well
     * @param label description field for document name or something like that
     * @return the job created by the upload
     */
    public FCproJob upload(File file, String ext, String profile, List<String> additionals, String label) {
        FileInputStream ins = null;
        try {
            ins = new FileInputStream(file);
            return upload(ins, ext, profile, additionals, label);
        } catch (FileNotFoundException e) {
            throw new GenericFCproClientException(e);
        } finally {
            if (ins != null) {
                try {
                    ins.close();
                } catch (IOException e) {
                    throw new GenericFCproClientException(e);
                }
            }
        }
    }

    /**
     * Upload and create a job from an {@code InputStream}. It is the
     * responsibility of the caller to close the InputStream instance.
     *
     * @param ins the {@code InputStream}
     * @param ext the target transformation format as mimetype
     * @param profile the server profile to use
     * @param additionals a list of file formats to be created as well
     * @param label description field for the job, can be null
     * @return the job created by the upload
     */
    public FCproJob upload(final InputStream ins, String ext, String profile, List<String> additionals, String label) {
        try {
            DefaultHttpClient httpclient = createHttpClient();

            // do api key check first
            validateHash();

            String url = endpoint + "/UploadJobEx3?ext=" + URLEncoder.encode(ext, urlEncodingCharset) + "&settingsName="
                    + URLEncoder.encode(profile, urlEncodingCharset) + "&maxPage=0&outExts=";
            if (additionals != null && !additionals.isEmpty()) {
                url = url + URLEncoder.encode(StringUtils.join(additionals, ","), urlEncodingCharset);
            }
            if (label != null) {
                url = url + "&label=" + URLEncoder.encode(label, urlEncodingCharset);

            }
            if (logger.isDebugEnabled()) {
                logger.debug("PUT url=" + url);
            }
            final HttpPut put = new HttpPut(url);
            final String printurl = url;
            put.addHeader("Content-Type", "application/octet-stream");
            ContentProducer cp = new ContentProducer() {
                @Override
                public void writeTo(OutputStream outstream) throws IOException {
                    long bytesWritten = IOUtils.copyLarge(ins, outstream);
                    if (logger.isDebugEnabled()) {
                        logger.debug("bytes transferred: " + bytesWritten + ", url=" + printurl);
                    }
                }
            };

            HttpEntity entity = new EntityTemplate(cp);
            put.setEntity(entity);
            prepareRequest(put);
            String content = httpclient.execute(put, new StringResponseHandler());
            if (logger.isDebugEnabled()) {
                logger.debug(content);
            }
            JSONObject json = new JSONObject(content);
            FCproJob job = jsonToJob(json.getJSONObject("UploadJobEx3Result"));
            return job;
        } catch (ClientProtocolException e) {
            throw new GenericFCproClientException(e);
        } catch (IOException | JSONException e) {
            throw new GenericFCproClientException(e);
        }

    }

    /**
     * Download transformation to a file. Single transformation version, always
     * downloads the first transformation.
     *
     * @param jobGuid the job GUID
     * @param outFile the file to save the transformation result to
     */
    public void download(String jobGuid, File outFile) {
        download(jobGuid, outFile, 0);
    }

    /**
     * Download transformation to a file.
     *
     * @param jobGuid the job GUID
     * @param outFile the file to save the transformation result to
     * @param index index of file in the transformation list
     */
    public void download(String jobGuid, File outFile, int index) {
        FileOutputStream outs = null;
        try {
            outs = new FileOutputStream(outFile);
            download(jobGuid, outs, index);
        } catch (FileNotFoundException e) {
            throw new GenericFCproClientException(e);
        } finally {
            if (outs != null) {
                try {
                    outs.close();
                } catch (IOException e) {
                    throw new GenericFCproClientException(e);
                }
            }
        }

    }

    /**
     * Downloads a result to the given {@link OutputStream}. The caller is
     * responsible to close the stream.
     *
     * @param jobGuid the job GUID
     * @param outs the {@code OutputStream}
     * @param index index of file in the transformation list
     */
    public void download(String jobGuid, OutputStream outs, int index) {

        DefaultHttpClient httpclient = createHttpClient();
        try {
            HttpGet get = new HttpGet(endpoint + "/GetResultEx?jobID=" + URLEncoder.encode(jobGuid, urlEncodingCharset) + "&index=" + index);
            if (logger.isDebugEnabled()) {
                logger.debug(get.getURI());
            }
            StreamResponseHandler handler = new StreamResponseHandler(outs);
            prepareRequest(get);
            httpclient.execute(get, handler);
        } catch (UnsupportedEncodingException e) {
            throw new GenericFCproClientException(e);
        } catch (ClientProtocolException e) {
            throw new GenericFCproClientException(e);
        } catch (IOException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Unmarshals a JSON encoded Job to a {@link FCproJob}.
     *
     * @param jsonJob the JSON object
     * @return the job
     */
    private FCproJob jsonToJob(JSONObject jsonJob) {
        FCproJob job = new FCproJob();
        try {
            job.setGuid(jsonJob.getString("JobGuid"));
            job.setJobId(jsonJob.getInt("JobID"));
            job.setPageCount(jsonJob.getInt("PageCount"));
            job.setStatus(FCproStatus.getStatusFromValue(jsonJob.getInt("Status")));
            job.setOwner(jsonJob.getString("UserName"));
            job.setLabel(jsonJob.getString("JobLabel"));
            job.setError(jsonJob.getString("Errors"));
            job.setCreated(parseISODate(jsonJob.getString("CreationDateISO")));
            job.setConverted(parseISODate(jsonJob.getString("FinishedDateISO")));
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
        return job;
    }

//    /*
//     * "CreationDate":"/Date(1359047825820+0100)/"
//     * "FinishedDate":"/Date(1359047538843+0100)/"
//     */
//    private Date parseTicksToDate(String input) {
//        final long TICKS_AT_EPOCH = 621355968000000000L;
//        final long TICKS_PER_MILLISECOND = 10000;
//
//        try {
//            String ticksAsString = input.split("\\(")[1].split("\\+")[0];
//            Long ticksAsLong = Long.valueOf(ticksAsString);
//            Date date = new Date((ticksAsLong - TICKS_AT_EPOCH) / TICKS_PER_MILLISECOND);
//            return date;
//        } catch (Exception e) {
//
//            return null;
//        }
//    }
    private Date parseISODate(String input) {
        if (input == null || input.trim().equals("")) {
            return null;
        }
        StringBuilder sb = new StringBuilder();
        final int i = input.indexOf("+") - 1;
        sb.append(input.substring(0, i)).append("GMT").append(input.substring(i));
        try {
            return sdf.parse(sb.toString());
        } catch (ParseException e) {
            logger.error(e.getMessage());
            return new Date();
        }
    }

    /**
     * Unmarshal a JSON encoded EnumJobsResult to a @{link FCproEnumJobsResult}.
     *
     * @param jsonEnumJobResult JSON
     * @return
     */
    private FCproEnumJobsResult jsonToEnumJobResult(JSONObject jsonEnumJobResult) {
        FCproEnumJobsResult result = new FCproEnumJobsResult();
        try {
            result.setJobEnumID(jsonEnumJobResult.getInt("JobEnumID"));
            result.setPagesCount(jsonEnumJobResult.getInt("PagesCount"));
            result.setPageSize(jsonEnumJobResult.getInt("PageSize"));
            result.setTotalJobs(jsonEnumJobResult.getInt("TotalJobs"));
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        }
        return result;
    }

    /**
     * Factory method for the http client, handling SSL and authentication. The
     * http client is configured to accept any SSL certificate from any host.
     *
     * @return
     */
    @SuppressWarnings("deprecation")
    private DefaultHttpClient createHttpClient() {
        // todo: DefaultHttpClient is told to be thread safe and could be reused
        DefaultHttpClient httpclient = new DefaultHttpClient();

        if (isSecureEndpoint()) {
            TrustManager anyTrustManager = new AnyTrustManager();
            X509HostnameVerifier anyHostnameVerifier = new AnyHostnameVerifier();

            SSLContext sslcontext;
            try {
                sslcontext = SSLContext.getInstance("TLS");

                sslcontext.init(null, new TrustManager[]{anyTrustManager}, null);
            } catch (NoSuchAlgorithmException | KeyManagementException e) {
                throw new GenericFCproClientException(e);
            }
            SSLSocketFactory ssf = new SSLSocketFactory(sslcontext);
            ssf.setHostnameVerifier(anyHostnameVerifier);
            ClientConnectionManager ccm = httpclient.getConnectionManager();
            SchemeRegistry sr = ccm.getSchemeRegistry();
            sr.register(new Scheme("https", ssf, 443));
        }

        // with basic auth
        if (StringUtils.isNotBlank(user)) {
            CredentialsProvider credsProvider = new BasicCredentialsProvider();
            Credentials credentials = new UsernamePasswordCredentials(user, password);
            credsProvider.setCredentials(AuthScope.ANY, credentials);
            httpclient.setCredentialsProvider(credsProvider);

        }

        // connection establish and transfer timeout
        httpclient.getParams().setIntParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, connectionTimeout);
        httpclient.getParams().setIntParameter(CoreConnectionPNames.SO_TIMEOUT, connectionTimeout);
        // TODO: make configurable
        httpclient.getParams().setBooleanParameter(CoreConnectionPNames.STALE_CONNECTION_CHECK, true);
        return httpclient;
    }

    /**
     * Handles the case where some json is returned. The service endpoint is
     * automatically prefixed.
     *
     * @param methodUri Uri to query, without the endpoint
     * @return
     */
    private JSONObject doSimpleJsonGet(String methodUri) {

        try {
            if (needAPIKeyCheck(methodUri)) {
                validateHash();
            }
            DefaultHttpClient httpclient = createHttpClient();
            HttpGet get = new HttpGet(endpoint + methodUri);
            prepareRequest(get);
            String content = httpclient.execute(get, new StringResponseHandler());
            if (logger.isDebugEnabled()) {
                logger.debug("Response: uri=" + methodUri + " content=" + content);
            }

            JSONObject json = new JSONObject(content);
            return json;
        } catch (JSONException e) {
            throw new GenericFCproClientException(e);
        } catch (ClientProtocolException e) {
            throw new GenericFCproClientException(e);
        } catch (IOException e) {
            throw new GenericFCproClientException(e);
        }
    }

    private boolean needAPIKeyCheck(String uri) {
        if (uri.contains("/GetServerHash")) {
            return false;
        }
        if (uri.contains("/GetSettingsCollection")) {
            return false;
        }
        if (uri.contains("/Admin/IsStarted")) {
            return false;
        }
        if (uri.contains("/Admin/Start")) {
            return false;
        }
        return true;

    }

    /**
     * Check if an API key is valid.
     *
     * @param apikey the API key
     * @return true if the api key is valid, else false.
     */
    public boolean isApiKeyValid(int apikey) {
        int serverHash = getServerHash();
        return apikey == serverHash;
    }

    /**
     * This method will be used for the REST calls, to check access permission.
     * However, the check is done once until successful per cient instance.
     *
     * @throws ApiKeyInvalidException if the key is invalid
     */
    private void validateHash() {
        if (apiKeyValidatedSuccessfully) {
            return;
        }
        if (!isApiKeyValid(apiHash)) {
            throw new ApiKeyInvalidException("The REST API-Key is invalid: key=" + apiHash);
        }
        apiKeyValidatedSuccessfully = true;
    }

    /**
     * Adds Basic authentication headers for preemptive authentication. This
     * works around a problem with Alfrescos one-shot ContentReaders.
     */
    private void prepareRequest(HttpRequestBase request) {
        if (StringUtils.isNotBlank(user)) {
            String clean = user + ":" + password;
            String value;
            try {
                byte[] encoded = Base64.encodeBase64(clean.getBytes(credentialsCharset));
                String stringVal = new String(encoded, "US-ASCII");
                value = "Basic " + stringVal;
            } catch (UnsupportedEncodingException e) {
                throw new GenericFCproClientException(e);
            }
            BasicHeader header = new BasicHeader("Authorization", value);
            request.setHeader(header);
        }
        // to keep it simple and robust, no connection management ist done
        request.setHeader("Connection", "close");
    }

    /**
     * Getter.
     *
     * @return the endpoint
     */
    public final String getEndpoint() {
        return endpoint;
    }

    /**
     * Setter.
     *
     * @param endpoint the endpoint
     */
    public final void setEndpoint(String endpoint) {
        this.endpoint = endpoint;
    }

    /**
     * Getter.
     *
     * @return the user
     */
    public final String getUser() {
        return user;
    }

    /**
     * Setter.
     *
     * @param user the user
     */
    public final void setUser(String user) {
        this.user = user;
    }

    /**
     * Getter.
     *
     * @return the password
     */
    protected final String getPassword() {
        return password;
    }

    /**
     * Setter.
     *
     * @param password the password
     */
    public final void setPassword(String password) {
        this.password = password;
    }

    /**
     * Check if the connection to the AutoOCR Service is secured through SSL.
     *
     * @return {@code true}, if the connection is SSL secured
     */
    public boolean isSecureEndpoint() {
        if (StringUtils.isBlank(endpoint)) {
            return false;
        }
        try {
            URI url = new URI(endpoint);
            String scheme = url.getScheme();
            if (StringUtils.isNotBlank(scheme)) {
                return scheme.equalsIgnoreCase("https");
            } else {
                return false;
            }
        } catch (URISyntaxException e) {
            throw new GenericFCproClientException(e);
        }
    }

    /**
     * Getter.
     *
     * @return credentials charset
     */
    public final String getCredentialsCharset() {
        return credentialsCharset;
    }

    /**
     * Setter.
     *
     * @param credentialsCharset credentials charset
     */
    public final void setCredentialsCharset(String credentialsCharset) {
        this.credentialsCharset = credentialsCharset;
    }

    /**
     * Getter.
     *
     * @return URL encoding charset
     */
    public final String getUrlEncodingCharset() {
        return urlEncodingCharset;
    }

    /**
     * Setter.
     *
     * @param urlEncodingCharset URL encoding charset
     */
    public final void setUrlEncodingCharset(String urlEncodingCharset) {
        this.urlEncodingCharset = urlEncodingCharset;
    }

    /**
     * Getter.
     *
     * @return connection timeout
     */
    public final int getConnectionTimeout() {
        return connectionTimeout;
    }

    /**
     * Setter.
     *
     * @param connectionTimeout connection timeout
     */
    public final void setConnectionTimeout(int connectionTimeout) {
        this.connectionTimeout = connectionTimeout;
    }

    /**
     * Setter.
     *
     * @param apiHash API hash
     */
    public final void setApiHash(int apiHash) {
        this.apiHash = apiHash;
    }

}
